// -*- C++ -*-
/*
 * CL Open EVSE Firmware
 *
 * Copyright (c) 2013-2014 Sam C. Lin <lincomatic@gmail.com>
 *
 * This file is part of CL Open EVSE.

 * CL Open EVSE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.

 * CL Open EVSE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with CL Open EVSE; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.


 **** RAPI protocol ****

Fx - function
Sx - set parameter
Gx - get parameter

command formats
1. with XOR checksum (recommended)
$cc pp pp ...^xk\r
2. with additive checksum (legacy)
$cc pp pp ...*ck\r
3. no checksum (FOR TESTING ONLY! DON'T USE FOR APPS)
$cc pp pp ...\r

\r = carriage return = 13d = 0x0D
cc = 2-letter command
pp = parameters
xk = 2-hex-digit checksum - 8-bit XOR of all characters before '^'
ck = 2-hex-digit checksum - 8-bit sum of all characters before '*'


response format
$OK [optional parameters]\r - success

$NK [optional parameters]\r - failure

asynchronous messages
$ST state\r - EVSE state transition - sent whenever EVSE state changes
 state: EVCS_STATE_xxx
$WF mode\r - Request client WiFi mode
 mode: WIFI_MODE_XXX
 (currently very long press (10 sec) of menu btn on OpenEVSE will send WIFI_MODE_AP_DEFAULT

commands

FB color - set LCD backlight color
colors:
 OFF 0
 RED 1
 YELLOW 3
 GREEN 2
 TEAL 6
 BLUE 4
 VIOLET 5
 WHITE 7

 $FB 7*03 - set backlight to white
FD - disable EVSE
 $FD*AE
FE - enable EVSE
 $FE*AF
FP x y text - print text on lcd display
FR - reset EVSE
 $FR*BC
FS - sleep EVSE
 $FS*BD

S0 0|1 - set LCD type
 $S0 0*F7 = monochrome backlight
 $S0 1*F8 = RGB backlight
S1 yr mo day hr min sec - set clock (RTC) yr=2-digit year
S2 0|1 - disable/enable ammeter calibration mode - ammeter is read even when not charging
 $S2 0*F9
 $S2 1*FA
S3 cnt - set charge time limit to cnt*15 minutes (0=disable, max=255)
SA currentscalefactor currentoffset - set ammeter settings
SC amps - set current capacity
   if amps < minimum current capacity, will set to minimum and return $NK
   if amps > maximum current capacity, will set to maximum and return $NK
SD 0|1 - disable/enable diode check
 $SD 0*0B
 $SD 1*0C
SE 0|1 - disable/enable command echo
 $SE 0*0C
 $SE 1*0D
 use this for interactive terminal sessions with RAPI.
 RAPI will echo back characters as they are typed, and add a <LF> character
 after its replies. Valid only over a serial connection, DO NOT USE on I2C
SG 0|1 - disable/enable ground check
 $SG 0*0E
 $SG 1*0F
SH kWh - set cHarge limit to kWh
SK - set accumulated Wh (v1.0.3+)
 $SK 0*12 - set accumulated Wh to 0
SL 1|2|A  - set service level L1/L2/Auto
 $SL 1*14
 $SL 2*15
 $SL A*24
SO ambientthresh irthresh - set Overtemperature thresholds
 thresholds are in 10ths of a degree Celcius
SR 0|1 - disable/enable stuck relay check
 $SR 0*19
 $SR 1*1A
ST starthr startmin endhr endmin - set timer
 $ST 0 0 0 0*0B - cancel timer
SV 0|1 - disable/enable vent required
 $SV 0*1D
 $SV 1*1E

G3 - get time limit
 response: OK cnt
 cnt*15 = minutes
        = 0 = no time limit
GA - get ammeter settings
 response: OK currentscalefactor currentoffset
 $GA*AC
GC - get current capacity range in amps
 response: OK minamps maxamps
 $GC*AE
GE - get settings
 response: OK amps(decimal) flags(hex)
 $GE*B0
GF - get fault counters
 response: OK gfitripcnt nogndtripcnt stuckrelaytripcnt (all values hex)
 maximum trip count = 0xFF for any counter
 $GF*B1
GG - get charging current and voltage
 response: OK milliamps millivolts
 AMMETER must be defined in order to get amps, otherwise returns -1 amps
 VOLTMETER must be defined in order to get voltage, otherwise returns -1 volts
 $GG*B2
GH - get cHarge limit
 response: OK kWh
 kWh = 0 = no charge limit
GM - get voltMeter settings
 response: OK voltcalefactor voltoffset
 $GM^2E
GO get Overtemperature thresholds
 response: OK ambientthresh irthresh
 thresholds are in 10ths of a degree Celcius
 $GO^2C
GP - get temPerature (v1.0.3+)
 $GP*BB
 response: OK ds3231temp
 ds3231temp - temperature from DS3231 RTC
 all temperatures are in 10th's of a degree Celcius
 if any temperature sensor is not installed, its return value will be 0
GS - get state
 response: OK state elapsed
 state: EVCS_STATE_xxx
 elapsed: elapsed charge time in seconds (valid only when in state C)
 $GS*BE
GT - get time (RTC)
 response OK yr mo day hr min sec       yr=2-digit year
 $GT*BF
GU - get energy usage (v1.0.3+)
 $GU*C0
 response OK Wattseconds Whacc
 Wattseconds - Watt-seconds used this charging session, note you'll divide Wattseconds by 3600 to get Wh
 Whacc - total Wh accumulated over all charging sessions, note you'll divide Wh by 1000 to get kWh
GV - get version
 response: OK firmware_version protocol_version
 $GV*C1

 *
 */
#ifndef RAPI_PROC_H
#define RAPI_PROC_H

#include <Arduino.h>
#include <TimeLib.h>

#define SERIAL_BITRATE 250000
#define VERSION "SIM 0.2 13/05/20"

#define MYSERIAL Serial1
#define RAPIVER "CL 0.3"
#define ESRAPI_BUFLEN 30
#define ESRAPI_SOC '$'         // start of command character 
#define ESRAPI_EOC 0x0A        // end of command character (0x0A = line feed)
#define ESRAPI_MAX_ARGS 10

#define EVCS_BUS '1'                   // EIA485 bus number (see RAPI)
#define EVCS_ID  '1'                   // number on bus
#define EVCS_BUS_STR "1"               // EIA485 bus number (see RAPI)
#define EVCS_ID_STR  "1"               // number on bus
#define EVCS_STATE_UNKNOWN 0x00
#define EVCS_STATE_A       0x01 // vehicle state A 12V - not connected
#define EVCS_STATE_B       0x02 // vehicle state B 9V - connected, ready
#define EVCS_STATE_C       0x03 // vehicle state C 6V - charging
#define EVCS_STATE_D       0x04 // vehicle state D 3V - vent required
#define EVCS_FAULT_STATE_BEGIN EVCS_STATE_D
#define EVCS_STATE_DIODE_CHK_FAILED 0x05 // diode check failed
#define EVCS_STATE_NO_GROUND 0x07 //bad ground
#define EVCS_STATE_STUCK_RELAY 0x08 //stuck relay
#define EVCS_STATE_OVER_TEMPERATURE 0x0A // over temperature error shutdown
#define EVCS_FAULT_STATE_END EVCS_STATE_OVER_TEMPERATURE
#define EVCS_STATE_SLEEPING 0xfe // waiting for timer
#define EVCS_STATE_DISABLED 0xff // disabled

#define DEFAULT_CURRENT_CAPACITY 16 // current capacity in amps
#define MIN_CURRENT_CAPACITY 6 // minimum allowable current in amps
#define MAX_CURRENT_CAPACITY 32 // maximum current in amps (EVCS_ 32A)

extern byte sleep_flag;
extern byte evcs_state;
const byte PIN_DE_RE = 10;
const char RAPI_VER[] PROGMEM = RAPIVER;
const char VERSTR[] PROGMEM = VERSION;

typedef union union4b {
  int8_t i8;
  uint8_t u8;
  int16_t i16;
  uint16_t u16;
  int32_t i32;
  uint32_t u32;
  unsigned u;
  int i;
} UNION4B,*PUNION4B;

inline void GetVerStr(char *buf) { strcpy_P(buf,VERSTR); }

class EvseRapiProcessor {
  char buffer[ESRAPI_BUFLEN]; // input buffer
  char out_buffer[100]; // output buffer for serial monitor
  int8_t bufCnt; // # valid bytes in buffer
  int8_t cmdIdCnt; // needed to distinguish busnr and idnr
  char *tokens[ESRAPI_MAX_ARGS];
  int8_t tokenCnt;
  byte echo = 1;
  virtual int available() = 0;
  virtual int read() = 0;
  virtual int write(uint8_t u8) = 0;
  virtual int write(const char *str) = 0;
  void reset() {
    buffer[0] = 0;
    bufCnt = 0;
    cmdIdCnt = 0;
  }
  int tokenize();
  int processCmd();
  void response(uint8_t ok);
public:
  EvseRapiProcessor();
  int doCmd();
  void sendEvseState();
  void sendRFID_NUID();
  virtual void init();
  
  uint16_t m_wFlags = 0x1A;
  char g_sTmp[50];
  uint16_t NoGndTripCnt = 1;
  uint16_t StuckRelayTripCnt = 1;  
  long ElapsedChargeTime = 100;
  int g_flagAllowCharging;
  long g_WattSeconds = 100;
  long g_WattHours_accumulated = 100;
  long temp = 245;

  uint32_t time_limit = 0;
  uint32_t kWh_limit = 0;
  uint32_t current_capacity = 0;
  uint32_t diode_check = 0;
  uint32_t ground_check = 0;  
  uint32_t stuck_relay_check = 0;
  uint32_t charge_limit = 0;
  uint32_t temp_ambient_thresh = 0;
  uint32_t temp_ir_thresh = 0;


  
};

class EvseSerialRapiProcessor : public EvseRapiProcessor {
  const unsigned long delay_dere_micros = 10;
  int available() {    
    digitalWrite(PIN_DE_RE,LOW);
    delayMicroseconds(delay_dere_micros);
    return Serial1.available();
  }
  int read() {
	  digitalWrite(PIN_DE_RE,LOW);    
    delayMicroseconds(50000000UL/SERIAL_BITRATE);    
    return Serial1.read();
  }
  int write(const char *str) {
    digitalWrite(PIN_DE_RE,HIGH);
    delayMicroseconds(delay_dere_micros);
	  int numbytes = Serial1.write(str);
    for (int i=0;i < numbytes;i++) {
      delayMicroseconds(10000000UL/SERIAL_BITRATE);
    }
    delayMicroseconds(10000000UL/SERIAL_BITRATE);
	  digitalWrite(PIN_DE_RE,LOW);
	  return numbytes;
  }
  int write(uint8_t u8) {
	  digitalWrite(PIN_DE_RE,HIGH);
    delayMicroseconds(delay_dere_micros);
	  int numbytes = Serial1.write(u8);
    delayMicroseconds(20000000UL/SERIAL_BITRATE);        
	  digitalWrite(PIN_DE_RE,LOW);
	  return numbytes;
  }
public:
  EvseSerialRapiProcessor();
  void init();
};

void RapiInit();
void RapiDoCmd();
void RapiSendEvseState(uint8_t nodupe=1);
void RapiSendRFID_NUID(uint8_t nodupe=1);

#endif
